@@ -1,29 +1,43 @@
 import sys
 import logging
-import threading
 import time
 import asyncio
 import traceback
 
+from discord.utils import maybe_coroutine
 from discord.player import AudioPlayer
+
 from .voice_source import Silence
 
 log = logging.getLogger(__name__)
 
 class MusicPlayer(AudioPlayer):
-    def __init__(self, track, client, *, after=None):
-        super().__init__(track.source, client, after=after)
+    def __init__(self, track, client):
+        super().__init__(track.source, client)
         self._play_silence = False
         self.track = track
         self._silence = Silence()
         self._leaving = client._leaving
         self._done = client._done
+        self._error = client._on_error
+        
+        # Client playlist
+        self.playlist = client._playlist
+
+        # Replacing thread name
+        # MusicPlayer_{ChannelID}_{Track pos}_{Track hash}
+        self.name = 'MusicPlayer_%s_%s_%s' % (
+            client.channel.id,
+            self.playlist.get_pos_from_track(track),
+            id(track)
+        )
 
         # For play the next song after done playing
         self.next_song = client._play_next_song
 
-        # Used for self.soft_stop()
-        self._soft_stop = False
+        # pre-play and post-play next song
+        self.pre_func = client._pre_next
+        self.post_func = client._post_next
 
         # For set_source()
         self._lock = client._lock
@@ -35,9 +49,8 @@ class MusicPlayer(AudioPlayer):
             self._current_error = exc
             self.stop()
         finally:
-            if not self._soft_stop:
-                self.source.cleanup()
-            self._call_after()
+            if not self._leaving.is_set() and self._connected.is_set():
+                self._call_after()
 
     def _do_run(self):
         self.loops = 0
@@ -92,44 +105,59 @@ class MusicPlayer(AudioPlayer):
             delay = max(0, self.DELAY + (next_time - time.perf_counter()))
             time.sleep(delay)
 
-    def _call_after(self):
+    def _handle_error(self):
         error = self._current_error
 
-        # Check if MusicClient.stop() is called
-        if self._done.is_set():
-            if error:
-                msg = 'Exception in voice thread {}'.format(self.name)
-                log.exception(msg, exc_info=error)
-                print(msg, file=sys.stderr)
-                traceback.print_exception(type(error), error, error.__traceback__)
-            return
-
-        # Play the next song
-        fut = asyncio.run_coroutine_threadsafe(self.next_song(), self.client.loop)
-        track = fut.result()
-
-        if self.after is not None:
-            # Check if after function is coroutine or not
-            if asyncio.iscoroutinefunction(self.after):
-                fut = asyncio.run_coroutine_threadsafe(self.after(error, track), self.client.loop)
-                exc = fut.exception()
-                if exc:
-                    log.exception('Calling the after function failed.')
-                    exc.__context__ = error
-                    traceback.print_exception(type(exc), exc, exc.__traceback__)
-                return
-            try:
-                self.after(error, track)
-            except Exception as exc:
-                log.exception('Calling the after function failed.')
+        if self._error:
+            fut = asyncio.run_coroutine_threadsafe(maybe_coroutine(self._error, error), self.client.loop)
+            exc = fut.exception()
+            if exc:
+                log.exception('Calling on player error function failed.')
                 exc.__context__ = error
                 traceback.print_exception(type(exc), exc, exc.__traceback__)
         elif error:
-            msg = 'Exception in voice thread {}'.format(self.name)
+            msg = 'Exception in MusicPlayer thread {}'.format(self.name)
             log.exception(msg, exc_info=error)
             print(msg, file=sys.stderr)
             traceback.print_exception(type(error), error, error.__traceback__)
 
+    def _call_after(self):
+        # Check if MusicClient.stop() is called
+        if self._done.is_set():
+            self._handle_error()
+            return
+
+        # get the next track
+        track = self.playlist.get_next_track()
+
+        # Call pre-play next function
+        if self.pre_func is not None:
+            fut = asyncio.run_coroutine_threadsafe(maybe_coroutine(self.pre_func, track), self.client.loop)
+            exc = fut.exception()
+            if exc:
+                log.exception('Calling the pre-play next track function failed.')
+                exc.__context__ = exc
+                traceback.print_exception(type(exc), exc, exc.__traceback__)
+
+        # Play the next song
+        fut = asyncio.run_coroutine_threadsafe(self.next_song(track), self.client.loop)
+        exc = fut.exception()
+        if exc:
+            log.exception('Calling play next track failed.')
+            exc.__context__ = exc
+            traceback.print_exception(type(exc), exc, exc.__traceback__)
+
+        # Call post-play next function
+        if self.post_func is not None:
+            fut = asyncio.run_coroutine_threadsafe(maybe_coroutine(self.post_func, track), self.client.loop)
+            exc = fut.exception()
+            if exc:
+                log.exception('Calling the post-play next track function failed.')
+                exc.__context__ = exc
+                traceback.print_exception(type(exc), exc, exc.__traceback__)
+
+        self._handle_error()
+
     def pause(self, *, update_speaking=True, play_silence=True):
         self._play_silence = play_silence
         super().pause(update_speaking=update_speaking)
@@ -138,25 +166,8 @@ class MusicPlayer(AudioPlayer):
         self._play_silence = False
         super().resume(update_speaking=update_speaking)
 
-    def soft_stop(self):
-        """Stop the player but not the ``MusicSource``
-
-        `MusicSource` will be restarted from zero using `recreate()` method
-        
-        This will be used in:
-        - `MusicClient.play_track_from_pos()`
-        - `MusicClient.next_track()`
-        - `MusicClient.previous_track()`
-        """
-        self._soft_stop = True
-
-        # Stop the player
-        self.stop()
-
-        # Wait until it terminates
-        self.join()
-        
-        # Start from zero
+    def stop(self):
+        super().stop()
         self.source.recreate()
         
     def _set_source(self, source):
