@@ -1,10 +1,12 @@
 import asyncio
 import traceback
 import threading
+import os
 
 from typing import Callable, Any, Union
 from discord.voice_client import VoiceClient
-from discord import opus
+from .opus_encoder import get_opus_encoder
+from .equalizer import Equalizer
 from .playlist import Playlist
 from .track import Track
 from .player import MusicPlayer
@@ -16,6 +18,7 @@ from .utils.errors import (
     NotConnected
 )
 
+_OpusEncoder = get_opus_encoder(os.environ.get('OPUS_ENCODER'))
 __all__ = (
     'MusicClient',
 )
@@ -29,7 +32,7 @@ class MusicClient(VoiceClient):
 
     Warning
     --------
-    It is important to add parameter `cls` with value :class:`MusicClient` to :meth:`discord.VoiceChannel.connect`,
+    It is important to add parameter ``cls`` with value :class:`MusicClient` to :meth:`discord.VoiceChannel.connect`,
     otherwise you wont get these features. For example:
 
     .. code-block:: python3
@@ -43,7 +46,12 @@ class MusicClient(VoiceClient):
     """
     def __init__(self, client, channel):
         super().__init__(client, channel)
-        self._after = None
+        self._pre_next = None
+        self._post_next = None
+        self._eq = None
+        self._volume = None
+        self._on_disconnect = None
+        self._on_error = None
 
         # Will be used for _stop()
         self._done = asyncio.Event()
@@ -57,6 +65,45 @@ class MusicClient(VoiceClient):
         # Will be used for music controls
         self._lock = asyncio.Lock()
 
+    def on_disconnect(self, func: Callable[[], Any]):
+        """A decorator that register a callable function as hook when disconnected
+
+        Note
+        -----
+        The function must be coroutine / async, otherwise it will raise error.
+        The function must be not take any parameters.
+
+        Parameters
+        -----------
+        func
+            a callable function (must be a coroutine function)
+        
+        Raises
+        -------
+        TypeError
+            The function is not coroutine or async
+        """
+        if not asyncio.iscoroutinefunction(func):
+            raise TypeError('The function is not coroutine or async')
+        self._on_disconnect = func
+
+    def on_player_error(self, func: Callable[[Exception], Any]):
+        """A decorator that register a callable function as hook when player encountered error.
+
+        The function can be normal function or coroutine (async) function.
+
+        Parameters
+        -----------
+        func: Callable[[:class:`Exception`], Any]
+            a callable function (can be a coroutine function)
+        
+        Raises
+        -------
+        TypeError
+            The function is not coroutine or async
+        """
+        self._on_error = func
+
     async def on_voice_state_update(self, data):
         self.session_id = data['session_id']
         channel_id = data['channel_id']
@@ -69,6 +116,8 @@ class MusicClient(VoiceClient):
                 # We're being disconnected so cleanup
                 self._leaving.set()
                 await self.disconnect()
+                if self._on_disconnect is not None:
+                    await self._on_disconnect()
             else:
                 guild = self.guild
                 self.channel = channel_id and guild and guild.get_channel(int(channel_id))
@@ -94,6 +143,7 @@ class MusicClient(VoiceClient):
                 self.socket.close()
 
     async def disconnect(self, *, force=False):
+        """Disconnects this voice client from voice."""
         if not force:
             if not self.is_connected():
                 return
@@ -103,6 +153,13 @@ class MusicClient(VoiceClient):
             await self._disconnect()
 
     async def move_to(self, channel):
+        """Moves you to a different voice channel.
+
+        Parameters
+        -----------
+        channel: :class:`discord.VoiceChannel`
+            The channel to move to. Must be a voice channel.
+        """
         async with self._lock:
             await super().move_to(channel)
     
@@ -122,32 +179,61 @@ class MusicClient(VoiceClient):
 
     # Playback controls
 
-    async def _play_next_song(self):
+    @property
+    def playlist(self):
+        """:class:`Playlist`: Return current playlist"""
+        return self._playlist
+    
+    async def set_playlist(self, playlist: Playlist, stop_player: bool=False):
+        """Replace current playlist with given playlist
+        
+        Parameters
+        -----------
+        playlist: :class:`Playlist`
+            A playlist that want to be setted
+        stop_player: :class:`bool`
+            Stop the player when changing playlist (if playing) and play current track from new playlist.
+        
+        Raises
+        -------
+        TypeError
+            "playlist" parameter is not :class:`Playlist`
+        """
+        if stop_player:
+            try:
+                await self.stop()
+            except MusicNotPlaying:
+                pass
+        if not isinstance(playlist, Playlist):
+            raise TypeError('playlist must an Playlist not {0.__class__.__name__}'.format(playlist))
+
+        self._playlist = playlist
+        async with self._lock:
+            self._play(playlist.get_current_track())
+
+    async def _play_next_song(self, track):
         # If disconnected then do nothing.
         if not self.is_connected():
             return
         
         # Play the next song
         async with self._lock:
-            track = self._playlist.get_next_track()
             if track:
-                self._play(track, self._after)
+                self._play(track)
 
-        return track
+    def before_play_next(self, func: Callable[[Union[Track, None]], Any]):
+        """A decorator that register callable function (can be coroutine function) as a pre-play next track
+        
+        This is useful for checking a streamable url or any type of set up required.
 
-
-    def register_after_callback(self, func: Callable[[Union[Exception, None], Union[Track, None]], Any]):
-        """Register a callable function (can be coroutine function)
-        for callback after player has done playing and play the next song or error occured.
+        The ``func`` callable must accept 1 parameter :class:`Track` that is the next track 
+        that want to be played.
 
         Parameters
-        ------------
-        func: Callable[[Union[:class:`Exception`, None], Union[:class:`Track`, None]], Any]
-            a callable function (can be coroutine function) that accept 2 parameters: :class:`Exception`
-            and :class:`Track`.
-            That :class:`Exception` is for exception in the player (if error happened) and 
-            :class:`Track` is for the next audio track. 
-        
+        -----------
+        func: Callable[[Union[:class:`Track`, ``None``]], Any]
+            The callable function to register as pre-play next track.
+
         Raises
         --------
         TypeError
@@ -155,7 +241,29 @@ class MusicClient(VoiceClient):
         """
         if not callable(func):
             raise TypeError('Expected a callable, got %s' % type(func))
-        self._after = func
+        self._pre_next = func
+
+    def after_play_next(self, func: Callable[[Union[Track, None]], Any]):
+        """A decorator that register callable function (can be coroutine function) as a post-play next track
+        
+        This is useful for sending announcement or any type of clean up required.
+
+        The ``func`` callable must accept 1 parameter :class:`Track` that is the next track 
+        that want to be played.
+
+        Parameters
+        -----------
+        func: Callable[[Union[:class:`Track`, ``None``]], Any]
+            The callable function to register as post-play next track.
+
+        Raises
+        --------
+        TypeError
+            Not a callable function
+        """
+        if not callable(func):
+            raise TypeError('Expected a callable, got %s' % type(func))
+        self._post_next = func
 
     def add_track(self, track: Track):
         """Add a track to playlist
@@ -163,15 +271,29 @@ class MusicClient(VoiceClient):
         Parameters
         -----------
         track: :class:`Track`
-            Audio Track that we're gonna play.
+            Audio Track that we're gonna add to playlist.
         """
         self._playlist.add_track(track)
 
-    def _play(self, track, after):
+    def _play(self, track):
         if not self.encoder and not track.source.is_opus():
-            self.encoder = opus.Encoder()
+            self.encoder = _OpusEncoder()
 
-        self._player = MusicPlayer(track, self, after=after)
+        # Apply equalizer
+        if self._eq:
+            try:
+                track.source.set_equalizer(self._eq)
+            except NotImplementedError:
+                pass
+        
+        # Apply volume
+        if self._volume:
+            try:
+                track.source.set_volume(self._volume)
+            except NotImplementedError:
+                pass
+
+        self._player = MusicPlayer(track, self)
         self._player.start()
 
         # we are playing
@@ -202,11 +324,9 @@ class MusicClient(VoiceClient):
             raise TypeError('track must an Track not {0.__class__.__name__}'.format(track))
 
         async with self._lock:
-            if self.is_playing():
-                self._playlist.add_track(track)
-                return
-            self._playlist.add_track(track)
-            self._play(track, self._after)
+            self.add_track(track)
+            if not self.is_playing():
+                self._play(track)
 
     async def play_track_from_pos(self, pos: int):
         """Play track from given pos
@@ -226,20 +346,14 @@ class MusicClient(VoiceClient):
         if not self.is_connected():
             raise NotConnected('Not connected to voice.')
         async with self._lock:
-            self._soft_stop()
+            self._stop()
             track = self._playlist.jump_to_pos(pos)
-            self._play(track, self._after)
+            self._play(track)
 
     def _stop(self):
-        if self._player:
-            self._player.stop()
-            self._player.join()
-            self._player = None
-
-    def _soft_stop(self):
         if self._player:
             self._done.set()
-            self._player.soft_stop()
+            self._player.stop()
             self._player = None
 
     async def stop(self):
@@ -260,7 +374,7 @@ class MusicClient(VoiceClient):
 
         Parameters
         -----------
-        play_silence: :class:`bool` (default: `True`)
+        play_silence: :class:`bool`
             if `True` play silence audio.
 
         Raises
@@ -345,11 +459,11 @@ class MusicClient(VoiceClient):
         if not self.is_connected():
             raise NotConnected('Not connected to voice.')
         async with self._lock:
-            self._soft_stop()
+            self._stop()
             track = self._playlist.get_next_track()
             if track is None:
                 raise NoMoreSongs('no more songs in playlist')
-            self._play(track, self._after)
+            self._play(track)
 
     async def previous_track(self):
         """Play previous track
@@ -364,11 +478,11 @@ class MusicClient(VoiceClient):
         if not self.is_connected():
             raise NotConnected('Not connected to voice.')
         async with self._lock:
-            self._soft_stop()
+            self._stop()
             track = self._playlist.get_previous_track()
             if track is None:
                 raise NoMoreSongs('no more songs in playlist')
-            self._play(track, self._after)
+            self._play(track)
 
     async def remove_track(self, track: Track):
         """Remove a track and stop the player (if given track same as playing track)
@@ -429,6 +543,74 @@ class MusicClient(VoiceClient):
 
     # Track related
 
+    @property
+    def equalizer(self):
+        """Optional[:class:`Equalizer`]: Return current equalizer music source being played (if playing)."""
+        return self.source.equalizer if self.source else None
+
+    def set_equalizer(self, equalizer: Equalizer):
+        """Set equalizer for music source
+
+        Note
+        -----
+        This will apply equalizer to all music sources to this playlist.
+        music sources that don't support equalizer or volume adjust will be ignored.
+
+        Parameters
+        -----------
+        equalier: :class:`Equalizer`
+            Equalizer that want to be setted in music source
+
+        Raises
+        -------
+        MusicNotPlaying
+            Not playing any audio
+        MusicClientException
+            current music source does not support equalizer
+        """
+        if not self.is_playing():
+            raise MusicNotPlaying('Not playing any audio')
+        if self.is_playing():
+            try:
+                self.source.set_equalizer(equalizer)
+            except NotImplementedError:
+                pass
+        self._eq = equalizer
+
+    @property
+    def volume(self):
+        """Optional[:class:`float`]: Return current volume music source being played (if playing)."""
+        return self.source.volume if self.source else None
+
+    def set_volume(self, volume: float):
+        """Set volume for music source for this channel.
+
+        Note
+        -----
+        This will apply volume to all music sources to this playlist.
+        music sources that don't support equalizer or volume adjust will be ignored.
+        
+        Parameters
+        -----------
+        volume: :class:`float`
+            Volume that want to be setted in music source
+        
+        Raises
+        -------
+        MusicNotPlaying
+            Not playing any audio
+        MusicClientException
+            current music source does not support volume adjust
+        """
+        if not self.is_playing():
+            raise MusicNotPlaying('Not playing any audio')
+        if self.is_playing():
+            try:
+                self.source.set_volume(volume)
+            except NotImplementedError:
+                pass
+        self._volume = volume
+
     @property
     def source(self):
         """Optional[:class:`MusicSource`]: The audio source being played, if playing."""
