@@ -1,27 +1,16 @@
 import sys
 import threading
+from typing import Union
 import discord
 import logging
 import audioop
 import wave
 import io
-from io import IOBase
+
+from io import BufferedIOBase
 from discord.opus import _OpusStruct as OpusEncoder
 from ..utils.errors import EqualizerError, IllegalSeek
-
-# Try to import equalizer module
-try:
-    from ..equalizer import PCMEqualizer, Equalizer
-    EQ_OK = True
-except EqualizerError:
-    # If failed to import equalizer module
-    # Re-create Equalizer class with no methods
-    class Equalizer:
-        def convert(self):
-            raise NotImplementedError
-    class PCMEqualizer(Equalizer):
-        pass
-    EQ_OK = False
+from ..equalizer import Equalizer
 
 __all__ = (
     'MusicSource','Silence', 'RawPCMAudio',
@@ -36,6 +25,30 @@ class MusicSource(discord.AudioSource):
     seek, rewind, equalizer and volume built-in to AudioSource
     """
 
+    def __init__(self):
+        self.__volume__ = None
+        self.__equalier__ = None
+
+    def read(self):
+        """Reads 20ms worth of audio.
+
+        Subclasses must implement this.
+
+        If the audio is complete, then returning an empty
+        :class:`bytes` to signal this is the way to do so.
+
+        If :meth:`is_opus` method returns ``True``, then it must return
+        20ms worth of Opus encoded audio. Otherwise, it must be 20ms
+        worth of 16-bit 48KHz stereo PCM, which is about 3,840 bytes
+        per frame (20ms worth of audio).
+
+        Returns
+        --------
+        :class:`bytes`
+            A bytes like object that represents the PCM or Opus data.
+        """
+        raise NotImplementedError
+
     def recreate(self):
         """Recreate audio source, useful for next and previous playback"""
         raise NotImplementedError
@@ -89,18 +102,29 @@ class MusicSource(discord.AudioSource):
         """
         raise NotImplementedError()
 
-    def set_volume(self, volume: float):
+    @property
+    def volume(self):
+        """Optional[:class:`float`]: Return current volume"""
+        return self.__volume__
+
+    def set_volume(self, volume: Union[float, None]):
         """
-        Set volume in float percentage
+        Set volume in float percentage.
+        Set to ``None`` to disable volume adjust.
 
         For example, 0.5 = 50%, 1.5 = 150%
 
         Parameters
         -----------
-        volume: :class:`volume`
+        volume: :class:`float`
             Set volume to music source
         """
-        raise NotImplementedError()
+        self.__volume__ = volume
+
+    @property
+    def equalizer(self):
+        """Optional[:class:`Equalizer`]: Return current equalizer"""
+        return self.__equalier__
 
     def set_equalizer(self, equalizer: Equalizer=None):
         """
@@ -111,7 +135,7 @@ class MusicSource(discord.AudioSource):
         equalizer: :class:`Equalizer`
             Set equalizer to music source
         """
-        raise NotImplementedError()
+        self.__equalier__ = equalizer
 
 class Silence(MusicSource):
     def read(self):
@@ -128,84 +152,43 @@ class RawPCMAudio(MusicSource):
     ------------
     stream: :class:`io.BufferedIOBase`
         file-like object
-    volume: :class:`float` or :class:`NoneType` (Optional, default: `0.5`)
+    volume: :class:`float` or :class:`NoneType`
         Set initial volume for AudioSource
 
     Attributes
     -----------
-    stream: :term:`py:file object`
+    stream: :class:`io.BufferedIOBase`
         A file-like object that reads byte data representing raw PCM.
     """
     def __init__(
         self,
-        stream: IOBase,
-        volume: float=0.5,
+        stream: BufferedIOBase,
+        volume: float=None,
     ):
+        super().__init__()
         self.stream = stream
         self._durations = 0
-        self._eq = None
+        self._eq = None # type: Equalizer
         self._lock = threading.Lock()
         self._buffered_eq = None
-        self._volume = max(volume, 0.0)
+        self.set_volume(volume)
 
     def read(self):
-        while True:
-            if self._eq is not None:
-                # At this point if RawPCMAudio using PCMEqualizer,
-                # the equalizer cannot convert audio if duration is too small (ex: 20ms)
-                # they will reproduce noisy sound.
-                # So the RawPCMAudio must read audio data at least for 1 second
-                # and then equalize it and move it to buffered equalized audio data.
-
-                def equalize(self, result=False):
-                    # The source will read the stream at least 1 second duration
-                    data = self.stream.read(OpusEncoder.FRAME_SIZE * 50) # 1 second duration
-
-                    # Return "exhausted" bytes to prevent re-reading stream
-                    if not data:
-                        return io.BytesIO(b'exhausted')
-
-                    # And then convert / equalize it
-                    eq_data = self._eq.convert(data)
-
-                    # Store it in buffered equalized audio data
-                    buffered_eq = io.BytesIO(eq_data)
-
-                    if result:
-                        return buffered_eq
-
-                if self._buffered_eq is None:
-                    self._buffered_eq = equalize(self, True)
-
-                # Read the buffered equalized audio data
-                data = self._buffered_eq.read(OpusEncoder.FRAME_SIZE)
-
-                if not data:
-                    self._buffered_eq = None
-                    continue
-                elif len(data) != OpusEncoder.FRAME_SIZE:
-                    return b''
-                self._durations += 0.020 # 20ms
-                if self._volume is None:
-                    return data
-                else:
-                    return audioop.mul(data, 2, min(self._volume, 2.0))
+        with self._lock:
+            # Read equalized audio data
+            if self.equalizer is not None:
+                data = self.equalizer.read()
             else:
-                if self._buffered_eq is not None:
-                    # Make sure that buffered eq is exhausted
-                    data = self._buffered_eq.read(OpusEncoder.FRAME_SIZE)
-                    if not data:
-                        self._buffered_eq = None
-                        continue
-                else:
-                    data = self.stream.read(OpusEncoder.FRAME_SIZE)
-                if len(data) != OpusEncoder.FRAME_SIZE:
-                    return b''
-                self._durations += 0.020 # 20ms
-                if self._volume is None:
-                    return data
-                else:
-                    return audioop.mul(data, 2, min(self._volume, 2.0))
+                data = self.stream.read(OpusEncoder.FRAME_SIZE)
+
+            if len(data) != OpusEncoder.FRAME_SIZE:
+                return b''
+
+            # Change volume audio
+            if self.volume is None:
+                return data
+            else:
+                return audioop.mul(data, 2, min(self.volume, 2.0))
     
     def cleanup(self):
         self.stream.close()
@@ -222,16 +205,17 @@ class RawPCMAudio(MusicSource):
     def get_stream_durations(self):
         return self._durations
 
-    def set_volume(self, volume: float):
-        self._volume = max(volume, 0.0)
+    def set_volume(self, volume):
+        vol = max(volume, 0.0) if volume is not None else None
+        super().set_volume(vol)
 
-    def set_equalizer(self, eq: PCMEqualizer=None):
-        if not EQ_OK:
-            raise EqualizerError('pydub and scipy need to be installed in order to use equalizer')
+    def set_equalizer(self, eq: Equalizer=None):
         if eq is not None:
-            if not isinstance(eq, PCMEqualizer):
-                raise EqualizerError('{0.__class__.__name__} is not PCMEqualizer'.format(eq))
-        self._eq = eq
+            if not isinstance(eq, Equalizer):
+                raise EqualizerError('{0.__class__.__name__} is not Equalizer'.format(eq))
+        with self._lock:
+            eq.setup(self.stream)
+            super().set_equalizer(eq)
 
     # -------------------------------------------
     # Formula seek and rewind for PCM-based Audio
@@ -322,16 +306,20 @@ class WAVAudio(RawPCMAudio):
     """
     Represents WAV audio stream
 
-    stream: :class:`io.BufferedIOBase`
-        file-like object
-    volume: :class:`float` or :class:`NoneType` (Optional, default: `0.5`)
+    file: Union[:class:`str`, :class:`io.BufferedIOBase`]
+        valid file location or file-like object.
+    volume: :class:`float` or :class:`NoneType`
         Set initial volume for AudioSource
     kwargs:
         These parameters will be passed in :class:`RawPCMAudio`
 
     """
-    def __init__(self, stream: io.IOBase, volume: float=0.5, **kwargs):
+    def __init__(self, file: Union[str, io.BufferedIOBase], volume: float=None, **kwargs):
         # Check if this stream is wav format
+        if isinstance(file, str):
+            stream = open(file, 'rb')
+        else:
+            stream = file
         new_stream = self._check_wav(stream)
         super().__init__(new_stream, volume, **kwargs)
 
@@ -344,26 +332,27 @@ class WAVAudio(RawPCMAudio):
     def _check_wav(self, stream):
         converted = io.BytesIO()
         info = wave.open(stream, 'rb')
-        output = wave.open(converted, 'wb')
 
         # If one of wav specifications (16-bit 48KHz) 
         # doesn't meet then convert it
         if False in self._check_wav_spec(info):
+            output = wave.open(converted, 'wb')
             output.setnchannels(2)
             output.setsampwidth(2)
             output.setframerate(48000)
         else:
-            output.close()
+            output = None
             info.close()
             return stream
 
         # Read the old stream and write it to new stream
         data = stream.read()
-        output.writeframesraw(data)
+        if output:
+            output.writeframesraw(data)
+            output.close()
 
         # Close the wave file
         info.close()
-        output.close()
 
         # Jump to 0 pos
         converted.seek(0, 0)
